# -*- coding: utf-8 -*-
# Copyright 2018-2019 Streamlit Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Smooths out some differences between python 2 and 3. It is meant to
be called as follows:

from __future__ import print_function, division, unicode_literals, absolute_import
from streamlit.compatibility import setup_2_3_shims
setup_2_3_shims(globals())
"""

import sys
import os

_dict = dict


def setup_2_3_shims(caller_globals):
    """
    Meant to be called as follows:

    setup_2_3_shims(globals())

    And sets up a bunch of compatibility aliases to make python 2 more like
    python 3.
    """
    if is_running_py3():
        caller_globals['dict_types'] = (type({}),)
        caller_globals['string_types'] = (type(''),)
        caller_globals['native_dict'] = _dict
    else:
        # These are the symbols we will export to the calling package.
        export_symbols = []

        # Override basic types.
        native_dict = _dict
        from builtins import range, map, str, dict, object, zip, int
        export_symbols += [
            'range', 'map', 'str', 'dict', 'object', 'zip', 'int',
            'native_dict']

        # Oerride the open function.
        from io import open
        export_symbols += ['open']

        from six import string_types
        export_symbols += ['string_types']

        # Export these symbols to the calling function's symbol table.
        for symbol in export_symbols:
            caller_globals[symbol] = locals()[symbol]

        # Special Cases
        caller_globals['FileNotFoundError'] = IOError
        caller_globals['dict_types'] = (dict, type({}))

        # Before we can call future.stanard_library, we need to make sure we're not
        # overriding any of the packages that it monkey patches or this can cause
        # some screwyness.
        illegal_package_names = ['urllib', 'test', 'dbm']
        current_directory_files = os.listdir('.')
        for illegal_package_name in illegal_package_names:
            illegal_source_file = illegal_package_name + '.py'
            assert illegal_source_file not in current_directory_files, \
                'File "%s" overrides a built-in package name.' \
                ' Please rename it.' % illegal_source_file

        # Do a bunch of dark monkey patching magic.
        from future.standard_library import install_aliases
        install_aliases()


def is_running_py3():
    """Returns True iff we're running 3 or above."""
    return sys.version_info >= (3, 0)
